/**************************************************************************
*  Copyright (c) 2023 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <stdint.h>
#include "tal.h"
#include "tcts.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/* 
 * Some TASK variables like stack and task control block.
 */
static OS_STACK (LEDStack,    TASK_LED_STK_SIZE);
static OS_STACK (StartStack,  TASK_START_STK_SIZE);
static OS_STACK (Float1Stack, TASK_FLOAT1_STK_SIZE);
static OS_STACK (Float2Stack, TASK_FLOAT2_STK_SIZE);

static OS_TCB TCBLed;
static OS_TCB TCBStartTask;
static OS_TCB TCBFloat1;
static OS_TCB TCBFloat2;

static uint8_t LEDStatus = 0;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  LEDSet                                                               */
/*                                                                       */
/*  In    : bChannel                                                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void LEDSet (uint8_t bChannel)
{
   LEDStatus |= bChannel;
   neorv32_gpio_port_set(LEDStatus);
   
} /* LEDSet */

/*************************************************************************/
/*  LEDClear                                                             */
/*                                                                       */
/*  In    : bChannel                                                     */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void LEDClear (uint8_t bChannel)
{
   LEDStatus &= ~bChannel;
   neorv32_gpio_port_set(LEDStatus);
   
} /* LEDClear */

/*************************************************************************/
/*  LEDInit                                                              */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void LEDInit (void)
{
   LEDStatus = 0;
   neorv32_gpio_port_set(LEDStatus);
   
} /* LEDInit */

/*************************************************************************/
/*  HardwareInit                                                         */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void HardwareInit (void)
{
   int rc = 0;
   
   /*
    * Capture all exceptions and give debug info via UART
    * this is not required, but keeps us safe.
    */
   neorv32_rte_setup();

   /* Init UART */
   term_Start();

   /* Check if GPIO unit is implemented at all */
   if (0 == neorv32_gpio_available()) 
   {
      rc = -1;
      neorv32_uart0_puts("\r\nError! No GPIO unit synthesized!\r\n");
   }

   /* Check if GPTMR unit is implemented at all */
   if (0 == neorv32_gptmr_available())
   {
      rc = -1;
      neorv32_uart0_puts("\r\nERROR! General purpose timer not implemented!\r\n");
   }
   
   /* In case of no error, start to init out HW */
   if (0 == rc)
   {
      LEDInit();
   }
   else
   {
      /* Error */
      while (1)
      {
         __asm__ volatile ("nop");
      }
   }
   
} /* HardwareInit */

/*************************************************************************/
/*  Float1Task                                                           */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void Float1Task (void *p)
{
   float fa = 0.1f;
   
   (void)p;
   
   while(1)
   {
      fa = fa + 0.1f;
      
      OS_TimeDly(TASK_FLOAT1_DELAY_MS);
   }
   
} /* Float1Task */

/*************************************************************************/
/*  Float2Task                                                           */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void Float2Task (void *p)
{
   float fa = 0.7f;
   
   (void)p;
   
   while(1)
   {
      fa = fa + 0.7f;
      
      OS_TimeDly(TASK_FLOAT2_DELAY_MS);
   }
   
} /* Float2Task */

/*************************************************************************/
/*  LEDTask                                                              */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void LEDTask (void *p)
{
   (void)p;
   
   while(1)
   {
      LEDSet(1);
      OS_TimeDly(TASK_LED_DELAY_MS);
      LEDClear(1);
      OS_TimeDly(TASK_LED_DELAY_MS);
   }
   
} /* LEDTask */

/*************************************************************************/
/*  StartTask                                                            */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void StartTask (void *p)
{
   (void)p;

   /*
    * The StartTask will be used to start all other tasks in the system.
    * At the end the priority will be set to "IDLE" priority.
    */
    
   /*
    * Start the System ticker
    */
   OS_SysTickStart();    

   /* Create the LED task */
   OS_TaskCreate(&TCBLed, LEDTask, NULL, TASK_LED_PRIORITY,
                 LEDStack, sizeof(LEDStack), 
                 "LEDTask");

   /* Create the FLOAT1 task */
   OS_TaskCreate(&TCBFloat1, Float1Task, NULL, TASK_FLOAT1_PRIORITY,
                 Float1Stack, sizeof(Float1Stack), 
                 "Float1Task");
   
   /* Create the FLOAT2 task */
   OS_TaskCreate(&TCBFloat2, Float2Task, NULL, TASK_FLOAT2_PRIORITY,
                 Float2Stack, sizeof(Float2Stack), 
                 "Float2Task");
                
   OS_TaskChangePriority(TASK_START_PRIORITY_IDLE);

   while (1)
   {
      LEDSet(2);
      OS_TimeDly(TASK_START_DELAY_MS);
      LEDClear(2);
      OS_TimeDly(TASK_START_DELAY_MS);
   }
   
} /* StartTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
int main (void)
{
   /*
    * Before the application which is using the TinyCTS can be started, 
    * the hardware of the target must be initialized.
    */
   HardwareInit();

   /* Init the TinyCTS */   
   OS_Init();
   
   /* 
    * Create the StartTask.
    * The StartTask is the one and only task which 
    * will start all other tasks of the system.
    */
   OS_TaskCreate(&TCBStartTask, StartTask, NULL, TASK_START_PRIORITY,
                 StartStack, sizeof(StartStack), 
                 "StartTask");
   
   /* 
    * OSStart must be the last function here.
    *  
    * Fasten your seatbelt, engine will be started...
    */        
   OS_Start();
        
   /*
    * This return here make no sense.
    * But to prevent the compiler warning:
    *    "return type of 'main' is not 'int'
    * We use an int as return :-)
    */ 
   return(0);
} /* main */

/*** EOF ***/

